/*
 * A pipe opener resource.
 *
 * Copyright (C) 2021, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <ipc/cap_alloc.h>
#include <resource/resource_server.h>

#include <l4/cxx/exceptions>

#include "pipe_opener_resource.h"
#include "pipe_opener_server.h"
#include "pipe_pager.h"



/* Support for providing access to pipes. */

PipeOpenerResource::PipeOpenerResource(Memory *memory)
: _memory(memory)
{
}

ipc_server_default_config_type PipeOpenerResource::config()
{
    return config_PipeOpener;
}



/* Pipe opener interface methods. */

long PipeOpenerResource::pipe(offset_t size, l4_cap_idx_t *reader, l4_cap_idx_t *writer)
{
    /* Both endpoints will employ a common paging coordinator. */

    PipePaging *paging;

    try
    {
        paging = new PipePaging(_memory, size);
    }
    catch (L4::Out_of_memory const &e)
    {
        return -L4_ENOMEM;
    }
 
    /* Each endpoint will have its own pager. */

    long err = open_endpoint(paging, true, writer);

    if (err)
        return err;

    err = open_endpoint(paging, false, reader);

    /* Failure to open an endpoint should invalidate both, plus the paging
       object. Also, any active server thread would need to be cancelled. The
       freeing of the writer capability attempts to do this. */

    if (err)
    {
        ipc_cap_free_um(*writer);
        return err;
    }

    return L4_EOK;
}

/* Start the endpoint server in a new thread. If the thread does not start, the
   resource (including pager) will be finalised. */

long PipeOpenerResource::open_endpoint(PipePaging *paging, bool writing, l4_cap_idx_t *endpoint)
{
    PipePager *pager = new PipePager(paging, writing);

    return ResourceServer(pager).start_thread(endpoint);
}

// vim: tabstop=4 expandtab shiftwidth=4
